package pps.core.prediction.service;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.date.DateUtil;
import cn.hutool.core.text.CharSequenceUtil;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.github.pagehelper.PageHelper;
import com.github.pagehelper.PageInfo;
import pps.cloud.base.service.IBasePowerLineCloudService;
import pps.cloud.base.service.data.base_power_line.DynamicQueryBasePowerLineInput;
import pps.cloud.base.service.data.base_power_line.DynamicQueryBasePowerLineOutput;
import pps.cloud.system.service.SysOrganizationCloudService;
import pps.cloud.system.service.data.GetAllOuListByOuIdInput;
import pps.cloud.system.service.data.GetSysOrganizationViewOutput;
import pps.core.prediction.entity.EnergyConsumptionAnalysisEnt;
import pps.core.prediction.enums.BusinessError;
import pps.core.prediction.mapper.EnergyConsumptionAnalysisMapper;
import pps.core.prediction.service.data.energy_consumption_analysis.*;
import xstartup.annotation.XService;
import xstartup.annotation.XText;
import xstartup.base.XContext;
import xstartup.base.exception.XServiceException;
import xstartup.data.XListResult;
import xstartup.data.XPageResult;
import xstartup.data.XSingleResult;
import xstartup.feature.api.annotation.XApiAnonymous;
import xstartup.feature.api.annotation.XApiGet;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.*;
import java.util.stream.Collectors;

/**
 * 能耗分析模块
 *
 * @author ZWT
 * @date 2023/11/23
 */
@XText("能耗分析模块")
@XService
public class EnergyConsumptionAnalysisService {

    /**
     * 能耗分析--能耗概览
     *
     * @param context 上下文
     * @param input   输入
     * @return {@link XSingleResult}<{@link GetEnergyConsumptionAnalysisViewOutput}>
     */
    @XText("能耗分析--能耗概览")
    @XApiAnonymous
    @XApiGet
    public XSingleResult<GetEnergyConsumptionAnalysisViewOutput> getEnergyConsumptionAnalysisView(XContext context, QueryEnergyConsumptionAnalysisInput input) {
        GetEnergyConsumptionAnalysisViewOutput output = new GetEnergyConsumptionAnalysisViewOutput();
        List<DynamicQueryBasePowerLineOutput> lineList = this.getBasePowerLineListByOuIdList(context, this.getOrgIdsByPath(context, input.getOuId()));
        if (CollUtil.isEmpty(lineList)) {
            output.setPowerGeneration(BigDecimal.ZERO);
            output.setPowerConsumption(BigDecimal.ZERO);
            output.setAbsorptionRate(BigDecimal.ZERO);
        } else {
            List<String> lineIdList = lineList.stream()
                    .map(DynamicQueryBasePowerLineOutput::getId)
                    .collect(Collectors.toList());
            input.getBetweenDate();
            EnergyConsumptionAnalysisMapper mapper = context.getBean(EnergyConsumptionAnalysisMapper.class);
            EnergyConsumptionAnalysisEnt analysisEnt = mapper.selectOne(new QueryWrapper<EnergyConsumptionAnalysisEnt>()
                    .select("IFNULL( SUM( power_generation ), 0 ) AS power_generation",
                            "IFNULL( SUM( power_consumption ), 0 ) AS power_consumption")
                    .lambda()
                    .in(EnergyConsumptionAnalysisEnt::getLineId, lineIdList)
                    .between(EnergyConsumptionAnalysisEnt::getDataDate, input.getStartTime(), input.getEndTime())
            );
            output.setPowerGeneration(analysisEnt.getPowerGeneration());
            output.setPowerConsumption(analysisEnt.getPowerConsumption());
            output.setAbsorptionRate(this.getAbsorptionRate(analysisEnt.getPowerGeneration(), analysisEnt.getPowerConsumption()));
        }
        return XSingleResult.success(output);
    }

    /**
     * 能耗分析--绿电消纳/用电趋势
     *
     * @param context 上下文
     * @param input   输入
     * @return {@link XListResult}<{@link DynamicQueryEnergyConsumptionAnalysisViewOutput}>
     */
    @XText("能耗分析--绿电消纳/用电趋势")
    @XApiAnonymous
    @XApiGet
    public XListResult<DynamicQueryEnergyConsumptionAnalysisViewOutput> queryEnergyConsumptionAnalysisList(XContext context, QueryEnergyConsumptionAnalysisInput input) {
        List<DynamicQueryEnergyConsumptionAnalysisViewOutput> list = new ArrayList<>(16);
        List<DynamicQueryBasePowerLineOutput> lineList = this.getBasePowerLineListByOuIdList(context, this.getOrgIdsByPath(context, input.getOuId()));
        if (CollUtil.isNotEmpty(lineList)) {
            List<String> lineIdList = lineList.stream()
                    .map(DynamicQueryBasePowerLineOutput::getId)
                    .collect(Collectors.toList());
            input.getBetweenDate();
            EnergyConsumptionAnalysisMapper mapper = context.getBean(EnergyConsumptionAnalysisMapper.class);
            List<EnergyConsumptionAnalysisEnt> analysisEntList = mapper.selectList(new QueryWrapper<EnergyConsumptionAnalysisEnt>()
                    .select("IFNULL( SUM( power_generation ), 0 ) AS power_generation",
                            "IFNULL( SUM( power_consumption ), 0 ) AS power_consumption",
                            "data_date")
                    .lambda()
                    .in(EnergyConsumptionAnalysisEnt::getLineId, lineIdList)
                    .between(EnergyConsumptionAnalysisEnt::getDataDate, input.getStartTime(), input.getEndTime())
                    .groupBy(EnergyConsumptionAnalysisEnt::getDataDate)
                    .orderByAsc(EnergyConsumptionAnalysisEnt::getDataDate)
            );
            if (CollUtil.isNotEmpty(analysisEntList)) {
                for (EnergyConsumptionAnalysisEnt ent : analysisEntList) {
                    list.add(DynamicQueryEnergyConsumptionAnalysisViewOutput.builder()
                            .powerConsumption(ent.getPowerConsumption())
                            .powerGeneration(ent.getPowerGeneration())
                            .dataDate(ent.getDataDate())
                            .dateFormat(DateUtil.format(ent.getDataDate(), "MM-dd"))
                            .absorptionRate(this.getAbsorptionRate(ent.getPowerGeneration(), ent.getPowerConsumption()))
                            .build());
                }
            }
        }
        return XListResult.success(list);
    }

    /**
     * 能耗分析--用电详情
     *
     * @param context 上下文
     * @param input   输入
     * @return {@link XPageResult}<{@link QueryEnergyConsumptionAnalysisViewOutput}>
     */
    @XText("能耗分析--用电详情")
    @XApiAnonymous
    @XApiGet
    public XPageResult<QueryEnergyConsumptionAnalysisViewOutput> queryEnergyConsumptionAnalysisView(XContext context, QueryEnergyConsumptionAnalysisPageInput input) {
        List<QueryEnergyConsumptionAnalysisViewOutput> outputs = new ArrayList<>(input.getLimit());
        List<EnergyConsumptionAnalysisEnt> list;
        List<DynamicQueryBasePowerLineOutput> lineList = this.getBasePowerLineListByOuIdList(context, this.getOrgIdsByPath(context, input.getOuId()));
        if (CollUtil.isNotEmpty(lineList)) {
            Map<String, String> lineMap = lineList.stream()
                    .collect(Collectors.toMap(DynamicQueryBasePowerLineOutput::getId, DynamicQueryBasePowerLineOutput::getLineName));
            Set<String> lineIds = lineMap.keySet();
            PageHelper.startPage(input.getPage(), input.getLimit());
            input.getBetweenDate();
            EnergyConsumptionAnalysisMapper mapper = context.getBean(EnergyConsumptionAnalysisMapper.class);
            Date startTime = input.getStartTime();
            Date endTime = input.getEndTime();
            //查本期数据
            list = mapper.selectList(this.getQueryWrapper4Page(lineIds, startTime, endTime));
            if (CollUtil.isNotEmpty(list)) {
                PageHelper.clearPage();
                //查上期数据
                switch (input.getDateType()) {
                    case 1:
                        endTime = DateUtil.offsetDay(startTime, -1);
                        startTime = DateUtil.offsetDay(startTime, -7);
                        break;
                    case 2:
                        endTime = DateUtil.offsetDay(startTime, -1);
                        startTime = DateUtil.offsetDay(startTime, -30);
                        break;
                    case 3:
                        int betweenDay = (int) DateUtil.betweenDay(startTime, endTime, true) + 1;
                        endTime = DateUtil.offsetDay(startTime, -1);
                        startTime = DateUtil.offsetDay(startTime, -betweenDay);
                        break;
                    default:
                        throw new XServiceException(BusinessError.DateTypeError);
                }
                List<EnergyConsumptionAnalysisEnt> periodList = mapper.selectList(this.getQueryWrapper4Page(lineIds, startTime, endTime));
                if (CollUtil.isNotEmpty(periodList)) {
                    for (EnergyConsumptionAnalysisEnt ent : list) {
                        EnergyConsumptionAnalysisEnt period = Optional.ofNullable(CollUtil.findOne(periodList, x -> CharSequenceUtil.equals(ent.getLineId(), x.getLineId())))
                                .orElse(EnergyConsumptionAnalysisEnt.builder()
                                        .powerConsumption(BigDecimal.ZERO)
                                        .powerGeneration(BigDecimal.ZERO)
                                        .build());
                        outputs.add(QueryEnergyConsumptionAnalysisViewOutput.builder()
                                .lineId(ent.getLineId())
                                .lineName(lineMap.get(ent.getLineId()))
                                .powerGeneration(ent.getPowerGeneration())
                                .generationCycleRatio(this.getCycleRatio(ent.getPowerGeneration(), period.getPowerGeneration()))
                                .powerConsumption(ent.getPowerConsumption())
                                .consumptionCycleRatio(this.getCycleRatio(ent.getPowerConsumption(), period.getPowerConsumption()))
                                .absorptionRate(this.getAbsorptionRate(ent.getPowerGeneration(), ent.getPowerConsumption()))
                                .build());
                    }
                } else {
                    for (EnergyConsumptionAnalysisEnt ent : list) {
                        outputs.add(QueryEnergyConsumptionAnalysisViewOutput.builder()
                                .lineId(ent.getLineId())
                                .lineName(lineMap.get(ent.getLineId()))
                                .powerGeneration(ent.getPowerGeneration())
                                .generationCycleRatio(BigDecimal.ZERO)
                                .powerConsumption(ent.getPowerConsumption())
                                .consumptionCycleRatio(BigDecimal.ZERO)
                                .absorptionRate(this.getAbsorptionRate(ent.getPowerGeneration(), ent.getPowerConsumption()))
                                .build());
                    }
                }
            }
        } else {
            list = new ArrayList<>(0);
        }
        PageInfo<EnergyConsumptionAnalysisEnt> pageInfo = new PageInfo<>(list);
        return XPageResult.success(outputs, input, pageInfo.getTotal());
    }

    /*-----------------------------------private-----------------------------------*/

    /**
     * 分页功能封装查询条件
     *
     * @param lineIds   线路ID
     * @param startTime 开始时间
     * @param endTime   结束时间
     * @return {@link LambdaQueryWrapper}<{@link EnergyConsumptionAnalysisEnt}>
     */
    private LambdaQueryWrapper<EnergyConsumptionAnalysisEnt> getQueryWrapper4Page(Collection<String> lineIds, Date startTime, Date endTime) {
        return new QueryWrapper<EnergyConsumptionAnalysisEnt>()
                .select("IFNULL( SUM( power_generation ), 0 ) AS power_generation",
                        "IFNULL( SUM( power_consumption ), 0 ) AS power_consumption",
                        "line_id")
                .lambda()
                .in(EnergyConsumptionAnalysisEnt::getLineId, lineIds)
                .between(EnergyConsumptionAnalysisEnt::getDataDate, startTime, endTime)
                .groupBy(EnergyConsumptionAnalysisEnt::getLineId)
                .orderByAsc(EnergyConsumptionAnalysisEnt::getPowerGeneration);
    }

    /**
     * 获取消纳率
     * 消纳率:光伏发电/累计用电*100%
     *
     * @param powerGeneration  发电
     * @param powerConsumption 功耗
     * @return {@link BigDecimal}
     */
    private BigDecimal getAbsorptionRate(BigDecimal powerGeneration, BigDecimal powerConsumption) {
        BigDecimal absorptionRate;
        if (powerConsumption.compareTo(BigDecimal.ZERO) > 0) {
            absorptionRate = powerGeneration.divide(powerConsumption, 4, RoundingMode.HALF_UP)
                    .multiply(new BigDecimal(100));
        } else {
            absorptionRate = BigDecimal.ZERO;
        }
        return absorptionRate;
    }

    /**
     * 获取环比
     * (本期-上期) /上期*100%
     *
     * @param currentPeriod 本期
     * @param lastPeriod    上期
     * @return {@link BigDecimal}
     */
    private BigDecimal getCycleRatio(BigDecimal currentPeriod, BigDecimal lastPeriod) {
        BigDecimal cycleRatio;
        if (lastPeriod.compareTo(BigDecimal.ZERO) > 0) {
            cycleRatio = currentPeriod.subtract(lastPeriod)
                    .divide(lastPeriod, 4, RoundingMode.HALF_UP)
                    .multiply(new BigDecimal(100));
        } else {
            cycleRatio = BigDecimal.ZERO;
        }
        return cycleRatio;
    }

    /**
     * 获取所有path包含组织ID的组织
     *
     * @param context 上下文
     * @param ouId    ou id
     * @return {@link List}<{@link String}>
     */
    private List<String> getOrgIdsByPath(XContext context, String ouId) {
        SysOrganizationCloudService organizationCloudService = context.getBean(SysOrganizationCloudService.class);
        GetAllOuListByOuIdInput ouIdInput = new GetAllOuListByOuIdInput();
        ouIdInput.setOuId(ouId);
        XListResult<GetSysOrganizationViewOutput> allListByOuId = organizationCloudService.getAllListByOuId(context, ouIdInput);
        allListByOuId.throwIfFail();
        return allListByOuId.getResult().stream()
                .map(GetSysOrganizationViewOutput::getId)
                .collect(Collectors.toList());
    }

    /**
     * 查询组织列表下所有线路
     *
     * @param context  上下文
     * @param ouIdList ou-id列表
     * @return {@link List}<{@link DynamicQueryBasePowerLineOutput}>
     */
    private List<DynamicQueryBasePowerLineOutput> getBasePowerLineListByOuIdList(XContext context, List<String> ouIdList) {
        IBasePowerLineCloudService basePowerLineCloudService = context.getBean(IBasePowerLineCloudService.class);
        XListResult<DynamicQueryBasePowerLineOutput> lineList = basePowerLineCloudService.getBasePowerLineListByOuIdList(context, DynamicQueryBasePowerLineInput.builder()
                .ouIdList(ouIdList)
                .build());
        lineList.throwIfFail();
        return lineList.getResult();
    }
}
